
#include "world.h"

World::World() { }

World::~World() { }

void World::load(const std::string &path)
{
    printf("Loading scene %s...\n", path.c_str());

    Any scene;
    scene.load(path);

    // Read the model table
    debugAssert(scene.containsKey("models"));
    const Table<std::string, Any> models = scene["models"].table();

    // Dump it to stdout
    printf("%d model(s)\n", (int)models.size());
    for (int i = 0; i < (int)models.size(); ++i)
        printf("    %s\n", models.getKeys()[i].c_str());

    // Read the entity table
    debugAssert(scene.containsKey("entities"));
    const Table<std::string, Any> &entities = scene["entities"].table();

    // Parse entities
    printf("%d entities(s)\n", (int)entities.size());
    Array<Surface::Ref> geometry;
    for (int i = 0; i < (int)entities.size(); ++i)
    {
        std::string key = entities.getKeys()[i];
        Any e = entities[key];
        std::string type = e.name();

        printf("    %s (%s) ... ", key.c_str(), type.c_str());

        if (type == "Camera") 
        {
            AnyTableReader props(e);
            m_camera = Camera::create(type, props);

            printf("done\n");
        }
        else if (type == "Light")
        {
            AnyTableReader props(e);
            m_lights.append(Light::create(type, props));

            printf("done\n");
        }
        else if (type == "VisibleEntity")
        {
            const Table<std::string, Any> &props = e.table();

            // Read the model from disk
            ArticulatedModel::Ref model = 
                ArticulatedModel::create(models[props["model"]]);

            // Pose it in world space
            Vector3 pos = Vector3::zero();
            if (props.containsKey("position"))
                pos = Vector3(props["position"]);

            Array<Surface::Ref> posed;
            model->pose(posed, CFrame(pos));

            // Add it to the scene
            for (int i = 0; i < posed.size(); ++i)
                geometry.append(posed[i]);

            printf("done\n");
        }
        else
        {
            printf("ignored (unknown entity type)\n");
        }
    }    

    // Build bounding interval hierarchy for scene geometry
    CPUVertexArray cpuVertexArray;
    Array<Tri> triArray;
    
    Surface::getTris(geometry, cpuVertexArray, triArray);
    for (int i = 0; i < triArray.size(); ++i)
        triArray[i].material()->setStorage(MOVE_TO_CPU);

    TriTree::Settings s;
    s.algorithm = TriTree::MEAN_EXTENT;
    m_tris.setContents(triArray, cpuVertexArray, s);
}

void World::unload()
{
    m_tris.clear();
    m_lights.clear();
}

Array<Light::Ref> World::lights()
{
    return m_lights;
}

int World::numLights()
{
    return (int)m_lights.size();
}

Light::Ref World::light(int i)
{
    return m_lights[i];
}

Camera::Ref World::camera()
{
    return m_camera;
}

void World::intersect(const Ray &ray, float &dist, shared_ptr<Surfel> &surf)
{
    dist = INFINITY; // No max distance
    surf = m_tris.intersectRay(ray, dist);
}

bool World::lineOfSight(const Vector3 &beg, const Vector3 &end)
{
    Vector3 d = end - beg;
    float dist = d.length();
    Ray ray = Ray::fromOriginAndDirection(beg, d / dist);

    // exitOnAnyHit -- Find any intersection, not the first (early exit)
    // twoSidedTest -- Don't cull back-facing triangles from intersectino test
    static const bool exitOnAnyHit = true, twoSidedTest = true;

    Tri::Intersector intersector;
    return !m_tris.intersectRay(ray, intersector, dist, exitOnAnyHit, twoSidedTest);
}

