#ifndef WORLD_H
#define WORLD_H

#include <G3D/G3DAll.h>

/** Represents a static scene with triangle mesh geometry, multiple lights, and
  * an initial camera specification
  */
class World
{
public:
    World();
    virtual ~World();

    /** Loads the geometry, lights and camera from a scene file.
      * Fails an assert if anything goes wrong.
      *
      * @param path The file to load (*.scn.any)
      */
    void load(const std::string &path);

    /** Clears the contents of this world object
      * Geometry and lights are cleared. The camera is not affected.
      */
    void unload();

    /** Gets all the lights in the scene */
    Array<Light::Ref> lights();

    /** Gets the number of lights in the scene */
    int numLights();

    /** Gets the i'th light */
    Light::Ref light(int i);

    /** Gets the scene's camera */
    Camera::Ref camera();

    /** Finds the first point a ray intersects with this scene
      *
      * @param ray  The ray to intersect
      * @param dist The distance from the ray origin to the point of
      *             intersection
      * @param surf The surface at the point of intersection
      */
    void intersect(const Ray &ray, float &dist, shared_ptr<Surfel> &surf);

    /** Determines whether an object occludes the line of sight from beg to end
     *
      * @param beg  The starting point
      * @param end  The ending point
      * @return     True if there is no geometry in the scene between the
      *             starting point and the ending point
      */
    bool lineOfSight(const Vector3 &beg, const Vector3 &end);

private:
    TriTree             m_tris;     // The scene's geometry in world space
    Array<Light::Ref>   m_lights;   // The scene's light in world space
    Camera::Ref         m_camera;   // The scene's camera
};

#endif
